from typing import Optional, List
from abc import ABC, abstractmethod
import pandas as pd # type: ignore
from pandas.core.generic import NDFrame # type:ignore

class BuildBase(ABC):
    """
    Base Class for Building a model
    """

    def __init__(self, scoring: str, forecast_period: int, verbose: int):
        self.scoring = scoring
        self.forecast_period = forecast_period
        self.verbose = verbose
        self.model = None

        self.original_target_col: str = ""
        self.original_preds: List[str] = []


    @abstractmethod
    def fit(self, ts_df: pd.DataFrame, target_col: str, cv: Optional[int] = None) -> object:
        """
        Fits the model to the data

        :param ts_df The time series data to be used for fitting the model
        :type ts_df pd.DataFrame

        :param target_col The column name of the target time series that needs to be modeled.
        All other columns will be considered as exogenous variables (if applicable to method)
        :type target_col str

        :param cv: Number of folds to use for cross validation.
        Number of observations in the Validation set for each fold = forecast period
        If None, a single fold is used
        :type cv Optional[int]

        :rtype object
        """


    @abstractmethod
    def refit(self, ts_df: pd.DataFrame) -> object:
        """
        Refits an already trained model using a new dataset
        Useful when fitting to the full data after testing with cross validation
        :param ts_df The time series data to be used for fitting the model
        :type ts_df pd.DataFrame
        :rtype object
        """

    @abstractmethod
    def predict(
        self,
        testdata: Optional[pd.DataFrame]=None,
        forecast_period: Optional[int] = None,
        simple: bool = True) -> NDFrame:
        """
        Return the predictions
        :param testdata The test dataframe containing the exogenous varaiables to be used for predicton.
        :type testdata Optional[pd.DataFrame]
        :param forecast_period The number of periods to make a prediction for.
        :type forecast_period Optional[int]
        :param simple If True, this method just returns the predictions.
        If False, it will return the standard error, lower and upper confidence interval (if available)
        :type simple bool
        :rtype NDFrame
        """

    def check_model_built(self):
        if self.model is None:
            raise AttributeError(
                "You are trying to perform an operation that requires the model to have been fit."+
                "However the model has not been fit yet. Please fit the model once before you try this operation."
            )

    def get_num_folds_from_cv(self, cv):
        if cv is None:
            NFOLDS = 1
        else:
            NFOLDS = cv

        return NFOLDS



##########################################################
#Defining AUTO_TIMESERIES here
##########################################################
import warnings
warnings.filterwarnings(action='ignore')
from typing import List, Dict, Optional, Tuple, Union

from datetime import datetime
import copy
from collections import defaultdict
import operator
from time import time
import pdb

# Tabular Data
import pandas as pd  # type: ignore
import numpy as np  # type: ignore

# Modeling
from sklearn.exceptions import DataConversionWarning # type: ignore
#### The warnings from Sklearn are so annoying that I have to shut it off ####
warnings.filterwarnings("ignore")
warnings.filterwarnings(action='ignore', category=DataConversionWarning)

# Plotting
import seaborn as sns  # type: ignore

def warn(*args, **kwargs):
    pass
warnings.warn = warn

############################################################
import seaborn as sns  # type: ignore
import matplotlib.pyplot as plt  # type: ignore
sns.set(style="white", color_codes=True)

#######################################
# Models
from .models import BuildBase, BuildArima, BuildAutoSarimax, BuildVAR, BuildML
from .models.build_prophet import BuildProphet


# Utils
from .utils import colorful, load_ts_data, convert_timeseries_dataframe_to_supervised, \
                   time_series_plot, print_static_rmse, print_dynamic_rmse, quick_ts_plot, \
                   test_stationarity, print_ts_model_stats


class auto_timeseries:
    def __init__(
        self,
        forecast_period: int = 5,
        score_type: str = 'rmse',
        time_interval: Optional[str] = None,
        non_seasonal_pdq: Optional[Tuple]=None,
        seasonality: bool = False,
        seasonal_period: int = 12,
        conf_int: float = 0.95,
        model_type: Union[str, List] = "stats",
        verbose: int = 0,
        *args,
        **kwargs
    ):
        """
        ####################################################################################
        ####                          Auto Time Series                                  ####
        ####         Developed by Ram Seshadri & Expanded by Nikhil Gupta               ####
        ####                        Python 3: 2018-2020                                 ####
        ####################################################################################
        Initialize an auto_timeseries object
        :score_type: The metric used for scoring the models. Default = 'rmse'
        Currently only 2 are supported:
        (1) RMSE
        (2) Normalized RMSE (ratio of RMSE to the standard deviation of actual)
        :type str

        :param time_interval Used to indicate the frequency at which the data is collected
        This is used for 2 purposes (1) in building the Prophet model and (2) used to impute
        the seasonal period for SARIMAX in case it is not provided by the user (None)
        Allowed values are:
          (1) 'months', 'month', 'm' for monthly frequency data
          (2) 'days', 'daily', 'd' for daily frequency data
          (3) 'weeks', 'weekly', 'w' for weekly frequency data
          (4) 'qtr', 'quarter', 'q' for quarterly frequency data
          (5) 'years', 'year', 'annual', 'y', 'a' for yearly frequency data
          (6) 'hours', 'hourly', 'h' for hourly frequency data
          (7) 'minutes', 'minute', 'min', 'n' for minute frequency data
          (8) 'seconds', 'second', 'sec', 's' for second frequency data
        :type time_interval Optional[str]

        :param: non_seasonal_pdq Indicates the maximum value of p, d, q to be used in the search for "stats" based models.
        If None, then the following values are assumed max_p = 3, max_d = 1, max_q = 3
        :type non_seasonal_pdq Optional[Tuple]

        :param seasonality Used in the building of the SARIMAX model only at this time.
        :type seasonality bool

        :param seasonal_period: Indicates the seasonality period in the data.
        Used in the building of the SARIMAX model only at this time.
        There is no impact of this argument if seasonality is set to False
        If None, the program will try to infer this from the time_interval (frequency) of the data
        (1) If frequency = Monthly, then seasonal_period = 12
        (1) If frequency = Daily, then seasonal_period = 30
        (1) If frequency = Weekly, then seasonal_period = 52
        (1) If frequency = Quarterly, then seasonal_period = 4
        (1) If frequency = Yearly, then seasonal_period = 1
        (1) If frequency = Hourly, then seasonal_period = 24
        (1) If frequency = Minutes, then seasonal_period = 60
        (1) If frequency = Seconds, then seasonal_period = 60
        :type seasonal_period int

        :param conf_int: Confidence Interval for building the Prophet model. Default: 0.95
        :type conf_int float

        :param model_type The type(s) of model to build. Default to building only statistical models
        Can be a string or a list of models. Allowed values are:
        'best', 'prophet', 'stats', 'SARIMAX', 'VAR', 'ML'.
        "prophet" will build a model using FB Prophet -> this means you must have FB Prophet installed
        "stats" will build statsmodels based SARIMAX and VAR models
        "ML" will build a machine learning model using Random Forests provided explanatory vars are given
        'best' will try to build all models and pick the best one
        If a list is provided, then only those models will be built
        WARNING: "best" might take some time for large data sets. We recommend that you
        choose a small sample from your data set bedfore attempting to run entire data.
        :type model_type: Union[str, List]

        :param verbose Indicates the verbosity of printing (Default = 0)
        :type verbose int

        ##################################################################################################
        AUTO_TIMESERIES IS A COMPLEX MODEL BUILDING UTILITY FOR TIME SERIES DATA. SINCE IT AUTOMATES MANY
        TASKS INVOLVED IN A COMPLEX ENDEAVOR, IT ASSUMES MANY INTELLIGENT DEFAULTS. BUT YOU CAN CHANGE THEM.
        Auto_Timeseries will rapidly build predictive models based on Statsmodels, Seasonal ARIMA
        and Scikit-Learn ML. It will automatically select the BEST model which gives best score specified.
        #####################################################################################################
        """
        self.ml_dict: Dict = {}
        self.score_type: str = score_type
        self.forecast_period =  forecast_period
        self.time_interval = time_interval
        self.non_seasonal_pdq = non_seasonal_pdq
        self.seasonality = seasonality
        self.seasonal_period = seasonal_period
        self.conf_int = conf_int
        if isinstance(model_type, str):
            model_type = [model_type]
        self.model_type = model_type
        self.verbose = verbose
        self.holidays = None
        self.growth = "linear"
        self.allowed_models = ['best', 'prophet', 'stats', 'ml', 'arima','ARIMA','Prophet','SARIMAX', 'VAR', 'ML']

        # new function.
        if args:
            for each_arg in args:
                print(each_arg)
        if kwargs:
            for key, value in zip(kwargs.keys(), kwargs.values()):
                if key == 'seasonal_PDQ':
                    print('seasonal_PDQ argument is deprecated. Please remove the argument in future.')
                if key == 'holidays':
                    print('holidays argument for FB Prophet given. It must be dictionary or DataFrame.')
                    self.holidays = value
                if key == 'growth':
                    print('growth argument of FB Prophet given. It must be "linear" or "logistic"')
                    self.growth = value

    def fit(
        self,
        traindata: Union[str, pd.DataFrame],
        ts_column: Union[str, int, List[str]],
        target: Union[str, List[str]],
        sep: Optional[str]=',', ## default is comma, string can be anything.
        cv: Optional[int]=5, ### Integer field cannot be defaulted to None
        ):
        """
        Train the auto_timeseries object
        # TODO: Complete docstring

        :param traindata Path for the data file or a dataframe. It accepts both.
        :type traindata Union[str, pd.DataFrame]

        :param ts_column Name of the datetime column in your dataset.
            If it is of type 'str', it will be treated as a column name.
            If it is of type 'int', it will be treated as the column number.
            If it is of type 'List', the first one will be picked and will be treated as the column name.
        :type ts_column Union[str, int, List[str]]

        :param target: Name of the column you are trying to predict. Target could also be the only column in your data

        :type target Union[str, List[str]]
            If it is of type 'str', it will be treated as a column name.
            If it is of type 'List', the first one will be picked and will be treated as the column name.

        :param cv Number of folds to use for cross validation.
            Number of observations in the Validation set for each fold = forecast period
            default is 5 fold cross validation.
        :type cv Optional[int]

        :param sep: Note that optionally you can give a separator for the data in your file.
            Default is None which treats the sep as a comma (datafile as a 'csv').
        :type sep Optional[str]
        """

        list_of_valid_time_ints = ['B','C','D','W','M','SM','BM','CBM',
                                        'MS','SMS','BMS','CBMS','Q','BQ','QS','BQS',
                                        'A,Y','BA,BY','AS,YS','BAS,BYS','BH',
                                        'H','T,min','S','L,ms','U,us','N']

        start = time()
        print("Start of Fit.....")

        ### first test the data for Stationary-ness #############
        test_stationarity(traindata[target].values, plot=False, verbose=True)

        ##### Best hyper-parameters in statsmodels chosen using the best aic, bic or whatever. Select here.
        stats_scoring = 'aic'

        ### If run_prophet is set to True, then only 1 model will be run and that is FB Prophet ##
        lag = copy.deepcopy(self.forecast_period)-1

        # if type(self.non_seasonal_pdq) == tuple:
        if isinstance(self.non_seasonal_pdq, tuple):
            p_max = self.non_seasonal_pdq[0]
            d_max = self.non_seasonal_pdq[1]
            q_max = self.non_seasonal_pdq[2]
        else:
            p_max = 3
            d_max = 1
            q_max = 3

        # Check 'ts_column' type
        if isinstance(ts_column, int):
            ### If ts_column is a number, then it means you need to convert it to a named variable
            # TODO: Fix this. ts_df is not defined.
            # You will need an argument to the load_ts_data function that will simply read the data
            # from the string location and return it as a dataframe as is
            ts_column = list(ts_df)[ts_column]
        if isinstance(ts_column, list):
            # If it is of type List, just pick the first one
            print("\nYou have provided a list as the 'ts_column' argument. Will pick the first value as the 'ts_column' name.")
            ts_column = ts_column[0]


        self.ts_column = ts_column

        # Check 'target' type
        if isinstance(target, list):
            target = target[0]
            print('    Auto_TS cannot handle Multi-Label targets. Taking first column in target list as Target = %s' %target)
        else:
            print('    Target variable given as = %s' %target)

        print("Start of loading of data.....")

        if sep is None:
            sep = ','

        ########## This is where we start the loading of the data file ######################
        if isinstance(traindata, str):
            if traindata != '':
                try:
                    ts_df = load_ts_data(traindata, self.ts_column, sep, target)
                    if isinstance(ts_df, str):
                        print("""Time Series column '%s' could not be converted to a Pandas date time column.
                            Please convert your ts_column into a pandas date-time and try again""" %self.ts_column)
                        return None
                    else:
                        print('    File loaded successfully. Shape of data set = %s' %(ts_df.shape,))
                except Exception:
                    print('File could not be loaded. Check the path or filename and try again')
                    return None
        elif isinstance(traindata, pd.DataFrame):
            print('Input is data frame. Performing Time Series Analysis')
            print(f"ts_column: {self.ts_column} sep: {sep} target: {target}")
            ts_df = load_ts_data(traindata, self.ts_column, sep, target)
            if isinstance(ts_df, str):
                print("""Time Series column '%s' could not be converted to a Pandas date time column.
                    Please convert your input into a date-time column  and try again""" %self.ts_column)
                return None
            else:
                print('    Dataframe loaded successfully. Shape of data set = %s' %(ts_df.shape,))
        else:
            print('File name is an empty string. Please check your input and try again')
            return None


        if ts_df.shape[1] == 1:
            ### If there is only one column, you assume that to be the target column ####
            target = list(ts_df)[0]


        preds = [x for x in list(ts_df) if x not in [self.ts_column, target]]

        try:
            time_series_plot(ts_df[target], lags=self.seasonal_period, title='Original Time Series',
                    chart_type='line', chart_freq=self.time_interval)
        except:
            print('Could not draw time series plot of data set. Continuing')
        ##################################################################################################
        ### Turn the time series index into a variable and calculate the difference.
        ### If the difference is not in days, then it is a hourly or minute based time series
        ### If the difference a multiple of days, then test it for weekly, monthly, quarterly, annual etc.
        ##################################################################################################
        if ts_df.index.dtype=='int' or ts_df.index.dtype=='float':
            ### You must convert the ts_df index into a date-time series using the ts_column given ####
            ts_df = ts_df.set_index(self.ts_column)
        ts_index = ts_df.index

        ## TODO: Be sure to also assign a frequency to the index column
        ## This will be helpful when finding the "future dataframe" especially for ARIMA, and ML.
        # https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases


        ##################    IF TIME INTERVAL IS NOT GIVEN DO THIS   ########################
        #### This is where the program tries to tease out the time period in the data set ####
        ######################################################################################
        if self.time_interval is None:
            print("Time Interval between obserations has not been provided. Auto_TS will try to infer this now...")
            ts_index = pd.to_datetime(ts_df.index)
            diff = (ts_index[1] - ts_index[0]).to_pytimedelta()
            diffdays = diff.days
            diffsecs = diff.seconds

            if diffsecs == 0:
                diff_in_hours = 0
                diff_in_days = abs(diffdays)
            else:
                diff_in_hours = abs(diffdays*24*3600 + diffsecs)/3600

            if diff_in_hours == 0 and diff_in_days >= 1:
                print('Time series input in days = %s' % diff_in_days)
                if diff_in_days == 7:
                    print('It is a Weekly time series.')
                    self.time_interval = 'weeks'
                elif diff_in_days == 1:
                    print('It is a Daily time series.')
                    self.time_interval = 'days'
                elif 28 <= diff_in_days < 89:
                    print('It is a Monthly time series.')
                    self.time_interval = 'months'
                elif 89 <= diff_in_days < 178:
                    print('It is a Quarterly time series.')
                    self.time_interval = 'qtr'
                elif 178 <= diff_in_days < 360:
                    print('It is a Semi Annual time series.')
                    self.time_interval = 'semi'
                elif diff_in_days >= 360:
                    print('It is an Annual time series.')
                    self.time_interval = 'years'
                else:
                    print('Time Series time delta is unknown')
                    return None
            if diff_in_days == 0:
                if diff_in_hours == 0:
                    print('Time series input in Minutes or Seconds = %s' % diff_in_hours)
                    print('It is a Minute time series.')
                    self.time_interval = 'minutes'
                elif diff_in_hours >= 1:
                    print('It is an Hourly time series.')
                    self.time_interval = 'hours'
                else:
                    print('It is an Unknown Time Series delta')
                    return None
        else:
            print('Time Interval is given as %s' % self.time_interval)
            if self.time_interval in list_of_valid_time_ints:
                print('    Correct Time interval given as a valid Pandas date-range frequency...')
            else:
                print('    Error: You must give a valid time interval frequency from Pandas date-range frequency codes')
                return None

        ################# This is where you test the data and find the time interval #######
        if self.time_interval is not None:
            if self.time_interval in list_of_valid_time_ints:
                pass
            else:
                self.time_interval = self.time_interval.strip().lower()
                if self.time_interval in ['months', 'month', 'm']:
                    self.time_interval = 'months'
                elif self.time_interval in ['days', 'daily', 'd']:
                    self.time_interval = 'days'
                elif self.time_interval in ['weeks', 'weekly', 'w']:
                    self.time_interval = 'weeks'
                elif self.time_interval in ['qtr', 'quarter', 'q']:
                    self.time_interval = 'qtr'
                elif self.time_interval in ['semi', 'semi-annual', '2q']:
                    self.time_interval = 'semi'
                elif self.time_interval in ['years', 'year', 'annual', 'y', 'a']:
                    self.time_interval = 'years'
                elif self.time_interval in ['hours', 'hourly', 'h']:
                    self.time_interval = 'hours'
                elif self.time_interval in ['minutes', 'minute', 'min', 'n']:
                    self.time_interval = 'minutes'
                elif self.time_interval in ['seconds', 'second', 'sec', 's']:
                    self.time_interval = 'seconds'
                else:
                    self.time_interval = 'months' # Default is Monthly
                    print('Time Interval not provided. Setting default as Monthly')
        else:
            print("(Error: 'self.time_interval' is None. This condition should not have occurred.")
            return

        # Impute seasonal_period if not provided by the user
        if self.seasonal_period is None:
            if self.time_interval == 'months':
                self.seasonal_period = 12
            elif self.time_interval == 'days':
                self.seasonal_period = 30
            elif self.time_interval in 'weeks':
                self.seasonal_period = 52
            elif self.time_interval in 'qtr':
                self.seasonal_period = 4
            elif self.time_interval in 'semi':
                self.seasonal_period = 2
            elif self.time_interval in 'years':
                self.seasonal_period = 1
            elif self.time_interval in 'hours':
                self.seasonal_period = 24
            elif self.time_interval in 'minutes':
                self.seasonal_period = 60
            elif self.time_interval in 'seconds':
                self.seasonal_period = 60
            else:
                self.seasonal_period = 12  # Default is Monthly


        ########################### This is where we store all models in a nested dictionary ##########
        mldict = lambda: defaultdict(mldict)
        self.ml_dict = mldict()
        try:
            if self.__any_contained_in_list(what_list=['best'], in_list=self.model_type):
                print(colorful.BOLD +'WARNING: Running best models will take time... Be Patient...' + colorful.END)
        except Exception:
            print('Check if your model type is a string or one of the available types of models')


        ######### This is when you need to use FB Prophet ###################################
        ### When the time interval given does not match the tested_time_interval, then use FB.
        #### Also when the number of rows in data set is very large, use FB Prophet, It is fast.
        #########                 FB Prophet              ###################################

        if self.__any_contained_in_list(what_list=['prophet', 'Prophet', 'best'], in_list=self.model_type):
            print("\n")
            print("="*50)
            print("Building Prophet Model")
            print("="*50)
            print("\n")

            name = 'Prophet'
            # Placeholder for cases when model can not be built
            score_val = np.inf
            model_build: Optional[BuildBase] = None
            model = None
            forecasts = None
            print(colorful.BOLD + '\nRunning Facebook Prophet Model...' + colorful.END)
            try:
                #### If FB prophet needs to run, it needs to be installed. Check it here ###
                model_build = BuildProphet(
                    self.forecast_period, self.time_interval,
                    self.score_type, self.verbose, self.conf_int, self.holidays, self.growth,
                    self.seasonality)
                model, forecast_df_folds, rmse_folds, norm_rmse_folds = model_build.fit(
                    ts_df=ts_df[[target]+preds],
                    target_col=target,
                    cv = cv,
                    time_col=self.ts_column)

                # forecasts = forecast_df['yhat'].values

                ##### Make sure that RMSE works, if not set it to np.inf  #########
                if self.score_type == 'rmse':
                    score_val = rmse_folds
                else:
                    score_val = norm_rmse_folds
            except Exception as e:
                print("Exception occurred while building Prophet model...")
                print(e)
                print('    FB Prophet may not be installed or Model is not running...')

            self.ml_dict[name]['model'] = model
            self.ml_dict[name]['forecast'] = forecast_df_folds
            self.ml_dict[name][self.score_type] = score_val
            self.ml_dict[name]['model_build'] = model_build


        # if self.__any_contained_in_list(what_list=['ARIMA', 'stats', 'best'], in_list=self.model_type):
        #     ################### Let's build an ARIMA Model and add results #################
        #     print("\n")
        #     print("="*50)
        #     print("Building ARIMA Model")
        #     print("="*50)
        #     print("\n")

        #     name = 'ARIMA'

        #     # Placeholder for cases when model can not be built
        #     score_val = np.inf
        #     model_build = None
        #     model = None
        #     forecasts = None
        #     print(colorful.BOLD + '\nRunning Non Seasonal ARIMA Model...' + colorful.END)
        #     try:
        #         model_build = BuildArima(
        #             stats_scoring, p_max, d_max, q_max,
        #             forecast_period=self.forecast_period, method='mle', verbose=self.verbose
        #         )
        #         model, forecasts, rmse, norm_rmse = model_build.fit(
        #             ts_df[target]
        #         )

        #         if self.score_type == 'rmse':
        #             score_val = rmse
        #         else:
        #             score_val = norm_rmse
        #     except Exception as e:
        #         print("Exception occurred while building ARIMA model...")
        #         print(e)
        #         print('    ARIMA model error: predictions not available.')

        #     self.ml_dict[name]['model'] = model
        #     self.ml_dict[name]['forecast'] = forecasts
        #     self.ml_dict[name][self.score_type] = score_val
        #     self.ml_dict[name]['model_build'] = model_build


        if self.__any_contained_in_list(what_list=['ARIMA','arima','auto_arima','auto_SARIMAX', 'stats', 'best'], in_list=self.model_type):
            ############# Let's build a SARIMAX Model and get results ########################
            print("\n")
            print("="*50)
            print("Building Auto SARIMAX Model")
            print("="*50)
            print("\n")

            name = 'auto_SARIMAX'
            # Placeholder for cases when model can not be built
            score_val = np.inf
            model_build = None
            model = None
            forecast_df_folds = None

            print(colorful.BOLD + '\nRunning Auto SARIMAX Model...' + colorful.END)
            try:
                model_build = BuildAutoSarimax(
                    scoring=stats_scoring,
                    seasonality=self.seasonality,
                    seasonal_period=self.seasonal_period,
                    p_max=p_max, d_max=d_max, q_max=q_max,
                    forecast_period=self.forecast_period,
                    verbose=self.verbose
                )
                model, forecast_df_folds, rmse_folds, norm_rmse_folds = model_build.fit(
                    ts_df=ts_df[[target]+preds],
                    target_col=target,
                    cv = cv
                )

                if self.score_type == 'rmse':
                    score_val = rmse_folds
                else:
                    score_val = norm_rmse_folds
            except Exception as e:
                print("Exception occurred while building Auto SARIMAX model...")
                print(e)
                print('    Auto SARIMAX model error: predictions not available.')


            self.ml_dict[name]['model'] = model
            self.ml_dict[name]['forecast'] = forecast_df_folds
            self.ml_dict[name][self.score_type] = score_val
            self.ml_dict[name]['model_build'] = model_build


        if self.__any_contained_in_list(what_list=['var','Var','VAR', 'stats', 'best'], in_list=self.model_type):
            ########### Let's build a VAR Model - but first we have to shift the predictor vars ####

            if ts_df.shape[0] > 1000 and self.__any_contained_in_list(what_list=['stats', 'best'], in_list=self.model_type):
                print("Skipping VAR Model since dataset is > 1000 rows and it will take too long")
            else:
                print("\n")
                print("="*50)
                print("Building VAR Model - best suited for small datasets < 1000 rows and < 10 columns")
                print("="*50)
                print("\n")

                name = 'VAR'
                # Placeholder for cases when model can not be built
                score_val = np.inf
                model_build = None
                model = None
                forecasts = None

                if len(preds) == 0:
                    print(colorful.BOLD + '\nNo VAR model created since no explanatory variables given in data set' + colorful.END)
                else:
                    try:
                        print(colorful.BOLD + '\nRunning VAR Model...' + colorful.END)
                        print('    Shifting %d predictors by 1 to align prior predictor values with current target values...'
                                                %len(preds))

                        # TODO: This causes an issue later in ML (most likely cause of https://github.com/AutoViML/Auto_TS/issues/15)
                        # Since we are passing ts_df there. Make sure you don't assign it
                        # back to the same variable. Make a copy and make changes to that copy.
                        ts_df_shifted = ts_df.copy(deep=True)
                        ts_df_shifted[preds] = ts_df_shifted[preds].shift(1)
                        ts_df_shifted.dropna(axis=0,inplace=True)

                        model_build = BuildVAR(scoring=stats_scoring, forecast_period=self.forecast_period, p_max=p_max, q_max=q_max)
                        model, forecasts, rmse, norm_rmse = model_build.fit(
                            ts_df_shifted[[target]+preds],
                            target_col=target,
                            cv = cv
                        )

                        if self.score_type == 'rmse':
                            score_val = rmse
                        else:
                            score_val = norm_rmse
                    except Exception as e:
                        print("Exception occurred while building VAR model...")
                        print(e)
                        warnings.warn('    VAR model error: predictions not available.')

                self.ml_dict[name]['model'] = model
                self.ml_dict[name]['forecast'] = forecasts
                self.ml_dict[name][self.score_type] = score_val
                self.ml_dict[name]['model_build'] = model_build

        if self.__any_contained_in_list(what_list=['ml', 'ML','best'], in_list=self.model_type):
            ########## Let's build a Machine Learning Model now with Time Series Data ################

            print("\n")
            print("="*50)
            print("Building ML Model")
            print("="*50)
            print("\n")

            name = 'ML'
            # Placeholder for cases when model can not be built
            score_val = np.inf
            model_build = None
            model = None
            forecasts = None

            if len(preds) == 0:
                print(colorful.BOLD + f'\nCreating lag={self.seasonal_period} variable using target for Machine Learning model...' + colorful.END)
                ### Set the lag to be 1 since we don't need too many lagged variables for univariate case
                self.lags = self.seasonal_period
                lag = self.seasonal_period
            else:
                print(colorful.BOLD + '\nRunning Machine Learning Models...' + colorful.END)
                #### Do not create excess lagged variables for ML model ##########
                if lag <= 4:
                    lag = 4 ### set the minimum lags to be at least 4 for ML models
                elif lag >= 10:
                    lag = 10 ### set the maximum lags to be not more than 10 for ML models
                print('    Shifting %d predictors by lag=%d to align prior predictor with current target...'
                            % (len(preds), lag))
            ####### Now make sure that there is only as few lags as needed ######
            model_build = BuildML(
                scoring=self.score_type,
                forecast_period = self.forecast_period,
                verbose=self.verbose)
            try:
                # best = model_build.fit(ts_df=ts_df, target_col=target, lags=lag)
                model, forecasts, rmse, norm_rmse = model_build.fit(
                    ts_df=ts_df,
                    target_col=target,
                    cv = cv,
                    lags=lag
                )

                if self.score_type == 'rmse':
                    score_val = rmse
                else:
                    score_val = norm_rmse
                # bestmodel = best[0]

                # #### Plotting actual vs predicted for ML Model #################
                # # TODO: Move inside the Build Class
                # plt.figure(figsize=(5, 5))
                # plt.scatter(train.append(test)[target].values,
                #             np.r_[bestmodel.predict(train[preds]), bestmodel.predict(test[preds])])
                # plt.xlabel('Actual')
                # plt.ylabel('Predicted')
                # plt.show(block=False)
                # ############ Draw a plot of the Time Series data ######
                # time_series_plot(dfxs[target], chart_time=self.time_interval)

            except Exception as e:
                print("Exception occurred while building ML model...")
                print(e)
                print('    For ML model, evaluation score is not available.')

            self.ml_dict[name]['model'] = model
            self.ml_dict[name]['forecast'] = forecasts
            self.ml_dict[name][self.score_type] = score_val
            self.ml_dict[name]['model_build'] = model_build

        if not self.__all_contained_in_list(what_list=self.model_type, in_list=self.allowed_models):
            print(f'The model_type should be any of the following: {self.allowed_models}. You entered {self.model_type}. Some models may not have been developed...')
            if len(list(self.ml_dict.keys())) == 0:
                return None

        ######## Selecting the best model based on the lowest rmse score ######
        best_model_name = self.get_best_model_name()
        print(colorful.BOLD + '\nBest Model is: ' + colorful.END + best_model_name)

        best_model_dict = self.ml_dict[best_model_name]
        if best_model_dict is not None:
            cv_scores = best_model_dict.get(self.score_type)
            mean_cv_score = self.__get_mean_cv_score(cv_scores)
        print("    Best Model (Mean CV) Score: %0.2f" % mean_cv_score) #self.ml_dict[best_model_name][self.score_type])

        end = time()
        elapsed = end-start
        print("\n\n" + "-"*50)
        print(f"Total time taken: {elapsed:.0f} seconds.")
        print("-"*50 + "\n\n")
        print("Leaderboard with best model on top of list:\n",self.get_leaderboard())
        return self

    def get_best_model_name(self) -> str:
        """
        Returns the best model name
        """
        f1_stats = {}
        for key, _ in self.ml_dict.items():
            cv_scores = self.ml_dict[key][self.score_type]

            # Standardize to a list
            if isinstance(cv_scores, np.ndarray):
                cv_scores = cv_scores.tolist()
            if not isinstance(cv_scores, List):
                cv_scores = [cv_scores]

            f1_stats[key] = sum(cv_scores)/len(cv_scores)

        best_model_name = min(f1_stats.items(), key=operator.itemgetter(1))[0]
        return best_model_name

    def get_best_model(self):
        """
        Returns the best model after training
        """
        return self.ml_dict.get(self.get_best_model_name()).get('model')

    def get_model(self, model_name: str):
        """
        Returns the specified model
        """
        if self.ml_dict.get(model_name) is not None:
            return self.ml_dict.get(model_name).get('model')
        else:
            print(f"Model with name '{model_name}' does not exist.")
            return None

    def get_best_model_build(self):
        """
        Returns the best model after training
        """
        return self.ml_dict.get(self.get_best_model_name()).get('model_build')

    def get_model_build(self, model_name: str):
        """
        Returns the specified model
        """
        if self.ml_dict.get(model_name) is not None:
            return self.ml_dict.get(model_name).get('model_build')
        else:
            print(f"Model with name '{model_name}' does not exist.")
            return None

    def get_ml_dict(self):
        """
        Returns the entire ML Dictionary
        """
        return self.ml_dict

    def predict(
        self,
        testdata,
        model: str = '',
        simple: bool = False,
        ):
        """
        Predict the results
        """

        if isinstance(model, str):
            if model == '':
                bestmodel = self.get_best_model_build()
            elif model.lower() == 'best':
                bestmodel = self.get_best_model_build()
            else:
                if self.get_model_build(model) is not None:
                    bestmodel = self.get_model_build(model)
                else:
                    print(f"(Error) Model of type '{model}' does not exist. No predictions will be made.")
                    return None
            self.model = bestmodel
        else:
            ### if no model is specified, just use the best model ###
            bestmodel = self.get_best_model_build()
            self.model = bestmodel


        if isinstance(testdata, pd.Series) or isinstance(testdata, pd.DataFrame):

            # During training, we internally converted a column datetime index to the dataframe date time index
            # We need to do the same while predicing for consistence
            if (model == 'ML') or self.get_best_model_name() == 'ML' or (model == 'best' and self.get_best_model_name() == 'ML'):
                if self.ts_column in testdata.columns:
                    testdata.set_index(self.ts_column, inplace=True)
                elif self.ts_column in testdata.index.name:
                    pass
                else:
                    print(f"(Error) Model to be used for prediction 'ML'. Hence, X_egogen' must have a column (or index) called '{self.ts_column}' corresponding to the original ts_index column passed during training. No predictions will be made.")
                    return None
                ### Now do the predictions using the final model asked to be predicted ###
                predictions = bestmodel.predict(testdata,simple=simple)
            elif model.lower() == 'prophet' or self.get_best_model_name() == 'Prophet':
                predictions = bestmodel.predict(testdata,simple=simple)
            elif self.get_best_model_name() == 'VAR':
                predictions = bestmodel.predict(testdata,simple=simple)
            else:
                predictions = bestmodel.predict(testdata,simple=simple)
        elif isinstance(testdata, int):
            #### if testdata is an Integer, then it appears to be a forecast period, then use it that way
            ### only certain stats-based models can use forecast period
            if (model == 'ML') or (model == 'best' and self.get_best_model_name() == 'ML'):
                print(f'{model} is an ML-based model, hence it cannot be used with a forecast period')
                predictions = None
            else:
                predictions = bestmodel.predict(testdata,simple=simple)
        else:
            ### if there is no testdata, at least they must give forecast_period
            if testdata is None:
                print('If test_data is None, then forecast_period must be given')
                return
        return predictions

    def get_leaderboard(self, ascending=True) -> Optional[pd.DataFrame]:
        """
        Returns the leaderboard after fitting
        """
        names = []
        mean_cv_scores = []
        # std_cv_scores = [] # TODO: Add later
        local_ml_dict = self.ml_dict
        if local_ml_dict is not None:
            model_names = list(local_ml_dict.keys())
            if model_names != []:
                for model_name in model_names:
                    names.append(model_name)
                    model_dict_single_model = local_ml_dict.get(model_name)
                    if model_dict_single_model is not None:
                        cv_scores = model_dict_single_model.get(self.score_type)
                        mean_cv_score = self.__get_mean_cv_score(cv_scores)
                        # if isinstance(cv_scores, float):
                        #     mean_cv_score = cv_scores
                        # else: # Assuming List
                        #     mean_cv_score = sum(cv_scores)/len(cv_scores)
                        mean_cv_scores.append(mean_cv_score)

                results = pd.DataFrame({"name": names, self.score_type: mean_cv_scores})
                results.sort_values(self.score_type, ascending=ascending, inplace=True)
                return results
            else:
                return None
        else:
            return None

    def plot_cv_scores(self, **kwargs):
        """
        Plots a boxplot of the cross validation scores for the various models.
        **kwargs: Keyword arguments to be passed to the seaborn boxplot call
        """
        cv_df = self.get_cv_scores()
        ax = sns.boxplot(x="Model", y="CV Scores", data=cv_df, **kwargs)
        return ax

    def get_cv_scores(self) -> pd.DataFrame:
        """
        Return a tidy data frame with the CV scores across all models
        :rtype pandas.DataFrame
        """
        cv_df = pd.DataFrame(columns=['Model', 'CV Scores'])
        for model in self.ml_dict.keys():
            rmses = self.ml_dict.get(model).get("rmse")
            new_row = {'Model':model, 'CV Scores':rmses}
            cv_df = cv_df.append(new_row, ignore_index=True)
            # print(f"Model: {model} RMSEs: {rmses}")
        cv_df = cv_df.explode('CV Scores').reset_index(drop=True)
        cv_df = cv_df.astype({"CV Scores": float})
        return cv_df

    def __get_mean_cv_score(self, cv_scores: Union[float, List]):
        """
        If gives a list fo cv scores, this will return the mean cv score
        If cv_score is a float (single value), it simply returns that
        """
        if isinstance(cv_scores, float):
            mean_cv_score = cv_scores
        else: # Assuming List
            mean_cv_score = sum(cv_scores)/len(cv_scores)
        return mean_cv_score

    def __any_contained_in_list(self, what_list: List[str], in_list: List[str], lower: bool = True) -> bool:
        """
        Returns True is any element in the 'in_list' is contained in the 'what_list'
        """
        if lower:
            what_list = [elem.lower() for elem in what_list]
            in_list = [elem.lower() for elem in in_list]

        return any([True if elem in in_list else False for elem in what_list])

    def __all_contained_in_list(self, what_list: List[str], in_list: List[str], lower: bool = True) -> bool:
        """
        Returns True is all elements in the 'in_list' are contained in the 'what_list'
        """
        if lower:
            what_list = [elem.lower() for elem in what_list]
            in_list = [elem.lower() for elem in in_list]

        return all([True if elem in in_list else False for elem in what_list])
#################################################################################
module_type = 'Running' if  __name__ == "__main__" else 'Imported'
version_number = '0.0.28'
print(f"""{module_type} auto_timeseries version:{version_number}. Call by using:
model = auto_timeseries(score_type='rmse', forecast_period=forecast_period,
                time_interval='Month',
                non_seasonal_pdq=None, seasonality=False, seasonal_period=12,
                model_type=['best'],
                verbose=2)

model.fit(traindata, ts_column,target)
model.predict(testdata, model='best')
""")
#################################################################################



import numpy as np  # type: ignore
import pandas as pd  # type: ignore
from pandas.core.generic import NDFrame # type:ignore

import matplotlib.pyplot as plt # type: ignore

from tscv import GapWalkForward # type: ignore

# imported SARIMAX from statsmodels pkg
from statsmodels.tsa.statespace.sarimax import SARIMAX  # type: ignore

from ..build_base import BuildBase
from .build_arima_base import BuildArimaBase

# helper functions
from ...utils import colorful, print_static_rmse, print_dynamic_rmse
from ...models.ar_based.param_finder import find_best_pdq_or_PDQ


# class BuildSarimax(BuildBase):
class BuildSarimax(BuildArimaBase):
    # def __init__(self, scoring, seasonality=False, seasonal_period=None, p_max=12, d_max=2, q_max=12, forecast_period=2, verbose=0):
    #     """
    #     Automatically build a SARIMAX Model
    #     """
    #     super().__init__(
    #         scoring=scoring,
    #         forecast_period=forecast_period,
    #         verbose=verbose
    #     )

    #     self.seasonality = seasonality
    #     self.seasonal_period = seasonal_period
    #     self.p_max = p_max
    #     self.d_max = d_max
    #     self.q_max = q_max

    #     self.best_p = None
    #     self.best_d = None
    #     self.best_q = None
    #     self.best_P = None
    #     self.best_D = None
    #     self.best_Q = None


    # def fit(self, ts_df: pd.DataFrame, target_col: str, cv: Optional[int]=None):
    #     """
    #     Build a Time Series Model using SARIMAX from statsmodels.
    #     """

    #     self.original_target_col = target_col
    #     self.original_preds = [x for x in list(ts_df) if x not in [self.original_target_col]]

    #     if len(self.original_preds) == 0:
    #         self.univariate = True
    #     else:
    #         self.univariate = False


    #     ##########################################
    #     #### Find best pdq and PDQ parameters ####
    #     ##########################################

    #     # NOTE: We use the entire dataset to compute the pdq and PDQ parameters.
    #     # Then we use the selected "best" parameters to check how well it
    #     # generalizes across the various folds (which may even be 1)

    #     # ## Added temporarily
    #     # ts_train = ts_df.iloc[:-self.forecast_period]
    #     # self.find_best_parameters(data = ts_train)
    #     self.find_best_parameters(data = ts_df)

    #     if self.verbose >= 1:
    #         print(f"\n\nBest Parameters:")
    #         print(f"p: {self.best_p}, d: {self.best_d}, q: {self.best_q}")
    #         print(f"P: {self.best_P}, D: {self.best_D}, Q: {self.best_Q}")
    #         print(f"Seasonality: {self.seasonality} Seasonal Period: {self.seasonal_period}")


    #     #######################################
    #     #### Cross Validation across Folds ####
    #     #######################################

    #     rmse_folds = []
    #     norm_rmse_folds = []
    #     forecast_df_folds = []

    #     NFOLDS = self.get_num_folds_from_cv(cv)
    #     cv = GapWalkForward(n_splits=NFOLDS, gap_size=0, test_size=self.forecast_period)
    #     for fold_number, (train, test) in enumerate(cv.split(ts_df)):
    #         ts_train = ts_df.iloc[train]
    #         ts_test = ts_df.iloc[test]

    #         if self.verbose >= 1:
    #             print(f"\n\nFold Number: {fold_number+1} --> Train Shape: {ts_train.shape} Test Shape: {ts_test.shape}")


    #         #########################################
    #         #### Define the model with fold data ####
    #         #########################################

    #         bestmodel = self.get_best_model(ts_train)

    #         ######################################
    #         #### Fit the model with fold data ####
    #         ######################################

    #         if self.verbose >= 1:
    #             print(colorful.BOLD + 'Fitting best SARIMAX model' + colorful.END)

    #         try:
    #             self.model = bestmodel.fit(disp=False)
    #             if self.verbose >= 1:
    #                 print('    Best %s metric = %0.1f' % (self.scoring, eval('self.model.' + self.scoring)))
    #         except Exception as e:
    #             print(e)
    #             print('Error: Getting Singular Matrix. Please try using other PDQ parameters or turn off Seasonality')
    #             return bestmodel, None, np.inf, np.inf

    #         if self.verbose >= 1:
    #             try:
    #                 self.model.plot_diagnostics(figsize=(16, 12))
    #             except:
    #                 print('Error: SARIMAX plot diagnostic. Continuing...')

    #         ### this is needed for static forecasts ####################
    #         # TODO: Check if this needs to be fixed to pick usimg self.original_target_col
    #         y_truth = ts_train[:]  #  TODO: Note that this is only univariate analysis

    #         if self.univariate:
    #             y_forecasted = self.model.predict(dynamic=False)
    #         else:
    #             y_forecasted = self.model.predict(dynamic=False, exog=ts_test[self.original_preds])

    #         concatenated = pd.concat([y_truth, y_forecasted], axis=1, keys=['original', 'predicted'])

    #         ### for SARIMAX, you don't have to restore differences since it predicts like actuals.###
    #         if self.verbose >= 1:
    #             print('Static Forecasts:')
    #             # Since you are differencing the data, some original data points will not be available
    #             # Hence taking from first available value.
    #             print_static_rmse(
    #                 concatenated['original'].values[self.best_d:],
    #                 concatenated['predicted'].values[self.best_d:],
    #                 verbose=self.verbose
    #             )

    #         ########### Dynamic One Step Ahead Forecast ###########################
    #         ### Dynamic Forecats are a better representation of true predictive power
    #         ## since they only use information from the time series up to a certain point,
    #         ## and after that, forecasts are generated using values from previous forecasted
    #         ## time points.
    #         #################################################################################
    #         # Now do dynamic forecast plotting for the last X steps of the data set ######

    #         if self.verbose >= 1:
    #             ax = concatenated[['original', 'predicted']][self.best_d:].plot(figsize=(16, 12))
    #             startdate = ts_df.index[-self.forecast_period-1]
    #             pred_dynamic = self.model.get_prediction(start=startdate, dynamic=True, full_results=True)
    #             pred_dynamic_ci = pred_dynamic.conf_int()
    #             pred_dynamic.predicted_mean.plot(label='Dynamic Forecast', ax=ax)
    #             try:
    #                 ax.fill_between(pred_dynamic_ci.index, pred_dynamic_ci.iloc[:, 0],
    #                                 pred_dynamic_ci.iloc[:, 1], color='k', alpha=.25)
    #                 ax.fill_betweenx(ax.get_ylim(), startdate, ts_train.index[-1], alpha=.1, zorder=-1)
    #             except:
    #                 pass
    #             ax.set_xlabel('Date')
    #             ax.set_ylabel('Levels')
    #             plt.legend()
    #             plt.show(block=False)

    #         # Extract the dynamic predicted and true values of our time series
    #         forecast_df = self.predict(testdata=ts_test[self.original_preds], simple=False)
    #         forecast_df_folds.append(forecast_df)

    #         # Extract Metrics
    #         if self.verbose >= 1:
    #             print('Dynamic %d-Period Forecast:' % (self.forecast_period))

    #         rmse, norm_rmse = print_dynamic_rmse(ts_test[self.original_target_col], forecast_df['mean'].values, ts_train[self.original_target_col], toprint=self.verbose)
    #         rmse_folds.append(rmse)
    #         norm_rmse_folds.append(norm_rmse)

    #         # TODO: Convert rmse_folds, rmse_norm_folds, forecasts_folds into base class attributes
    #         # TODO: Add gettes and seters for these class attributes.
    #         # This will ensure consistency across various model build types.


    #     # This is taking the std of entire dataset and using that to normalize
    #     # vs. other approach that was using std of individual folds to stansardize.
    #     # Technically this is not correct, but in order to do Apples:Aples compatison with ML
    #     # (sklearn) based cross_val_score, we need to do this since we dont get indicidual folds
    #     # back for cross_val_score. If at a later point in time, we can get this, then,
    #     # we can revert back to dividing by individual fold std values.
    #     norm_rmse_folds2 = rmse_folds/ts_df[self.original_target_col].values.std()  # Same as what was there in print_dynamic_rmse()

    #     # print(f"SARIMAX Norm RMSE (Original): {norm_rmse_folds}")
    #     # print(f"SARIMAX Norm RMSE (New): {norm_rmse_folds2}")

    #     ###############################################
    #     #### Refit the model on the entire dataset ####
    #     ###############################################
    #     self.refit(ts_df=ts_df)

    #     if self.verbose >= 1:
    #         print(self.model.summary())

    #     # return self.model, forecast_df_folds, rmse_folds, norm_rmse_folds
    #     return self.model, forecast_df_folds, rmse_folds, norm_rmse_folds2

    # def refit(self, ts_df: pd.DataFrame) -> object:
    #     """
    #     Refits an already trained model using a new dataset
    #     Useful when fitting to the full data after testing with cross validation
    #     :param ts_df The time series data to be used for fitting the model
    #     :type ts_df pd.DataFrame
    #     :rtype object
    #     """

    #     bestmodel = self.get_best_model(ts_df)

    #     print(colorful.BOLD + 'Refitting data with previously found best parameters' + colorful.END)
    #     try:
    #         self.model = bestmodel.fit(disp=False)
    #         print('    Best %s metric = %0.1f' % (self.scoring, eval('self.model.' + self.scoring)))
    #     except Exception as e:
    #         print(e)

    #     return self


    def find_best_parameters(self, data: pd.DataFrame):
        """
        Given a dataset, finds the best parameters using the settings in the class
        """

        if not self.seasonality:
            if self.verbose >= 1:
                print('Building a Non Seasonal Model...')
                print('\nFinding best Non Seasonal Parameters:')
            # TODO: Check if we need to also pass the exogenous variables here and
            # change the functionality of find_best_pdq_or_PDQ to incorporate these
            # exogenoug variables.
            self.best_p, self.best_d, self.best_q, best_bic, _ = find_best_pdq_or_PDQ(
                ts_df=data[self.original_target_col],
                scoring=self.scoring,
                p_max=self.p_max, d_max=self.d_max, q_max=self.q_max,
                non_seasonal_pdq=None,
                seasonal_period=None,
                seasonality=False,
                verbose=self.verbose
            )

            if self.verbose >= 1:
                print('\nBest model is: Non Seasonal SARIMAX(%d,%d,%d), %s = %0.3f' % (
                    self.best_p, self.best_d, self.best_q, self.scoring, best_bic))
        else:
            if self.verbose >= 1:
                print(colorful.BOLD + 'Building a Seasonal Model...'+colorful.END)
                print(colorful.BOLD + '\n    Finding best Non-Seasonal pdq Parameters:' + colorful.END)
            # TODO: Check if we need to also pass the exogenous variables here and
            # change the functionality of find_best_pdq_or_PDQ to incorporate these
            # exogenoug variables.
            self.best_p, self.best_d, self.best_q, _, _ = find_best_pdq_or_PDQ(
                ts_df=data[self.original_target_col],
                scoring=self.scoring,
                p_max=self.p_max, d_max=self.d_max, q_max=self.q_max,
                non_seasonal_pdq=None,  # we need to figure this out ...
                seasonal_period=None,
                seasonality=False,  # setting seasonality = False for p, d, q
                verbose=self.verbose
            )

            if self.verbose >= 1:
                print(colorful.BOLD + '\n    Finding best Seasonal PDQ Model Parameters:' + colorful.END)
            # TODO: Check if we need to also pass the exogenous variables here and
            # change the functionality of find_best_pdq_or_PDQ to incorporate these
            # exogenoug variables.
            self.best_P, self.best_D, self.best_Q, best_bic, self.seasonality = find_best_pdq_or_PDQ(
                ts_df=data[self.original_target_col],
                scoring=self.scoring,
                p_max=self.p_max, d_max=self.d_max, q_max=self.q_max,
                non_seasonal_pdq=(self.best_p, self.best_d, self.best_q), # found previously ...
                seasonal_period=self.seasonal_period,  # passing seasonal period
                seasonality=True,  # setting seasonality = True for P, D, Q
                verbose=self.verbose
            )

            if self.seasonality:
                if self.verbose >= 1:
                    print('\nBest model is a Seasonal SARIMAX(%d,%d,%d)*(%d,%d,%d,%d), %s = %0.3f' % (
                        self.best_p, self.best_d, self.best_q,
                        self.best_P, self.best_D, self.best_Q,
                        self.seasonal_period, self.scoring, best_bic))
            else:
                if self.verbose >= 1:
                    print('\nEven though seasonality has been set to True, the best model is a Non Seasonal SARIMAX(%d,%d,%d)' % (
                        self.best_p, self.best_d, self.best_q))





    # def get_best_model(self, data: pd.DataFrame):
    #     """
    #     Returns the 'unfit' SARIMAX model with the given dataset and the
    #     selected best parameters. This can be used to fit or refit the model.
    #     """

    #     # In order to get forecasts to be in the same value ranges of the orig_endogs, you
    #     # must  set the simple_differencing = False and the start_params to be the same as ARIMA.
    #     # That is the only way to ensure that the output of this model iscomparable to other ARIMA models

    #     if not self.seasonality:
    #         if self.univariate:
    #             bestmodel = SARIMAX(
    #                 endog=data[self.original_target_col],
    #                 # exog=data[self.original_preds],
    #                 order=(self.best_p, self.best_d, self.best_q),
    #                 enforce_stationarity=False,
    #                 enforce_invertibility=False,
    #                 trend='ct',
    #                 start_params=[0, 0, 0, 1],
    #                 simple_differencing=False)
    #         else:
    #             bestmodel = SARIMAX(
    #                 endog=data[self.original_target_col],
    #                 exog=data[self.original_preds],
    #                 order=(self.best_p, self.best_d, self.best_q),
    #                 enforce_stationarity=False,
    #                 enforce_invertibility=False,
    #                 trend='ct',
    #                 start_params=[0, 0, 0, 1],
    #                 simple_differencing=False)
    #     else:
    #         if self.univariate:
    #             bestmodel = SARIMAX(
    #                 endog=data[self.original_target_col],
    #                 # exog=data[self.original_preds],
    #                 order=(self.best_p, self.best_d, self.best_q),
    #                 seasonal_order=(self.best_P, self.best_D, self.best_Q, self.seasonal_period),
    #                 enforce_stationarity=False,
    #                 enforce_invertibility=False,
    #                 trend='ct',
    #                 start_params=[0, 0, 0, 1],
    #                 simple_differencing=False
    #             )
    #         else:
    #             bestmodel = SARIMAX(
    #                 endog=data[self.original_target_col],
    #                 exog=data[self.original_preds],
    #                 order=(self.best_p, self.best_d, self.best_q),
    #                 seasonal_order=(self.best_P, self.best_D, self.best_Q, self.seasonal_period),
    #                 enforce_stationarity=False,
    #                 enforce_invertibility=False,
    #                 trend='ct',
    #                 start_params=[0, 0, 0, 1],
    #                 simple_differencing=False
    #             )

    #     return bestmodel

    # def predict(
    #     self,
    #     testdata: Optional[pd.DataFrame]=None,
    #     forecast_period: Optional[int] = None,
    #     simple: bool = True) -> NDFrame:
    #     """
    #     Return the predictions
    #     """
    #     # Extract the dynamic predicted and true values of our time series

    #     if self.univariate:
    #         if forecast_period is None:
    #             # use the forecast period used during training
    #             forecast_period = self.forecast_period
    #     else:
    #         if testdata is None:
    #             raise ValueError("SARIMAX needs testdata to make predictions, but this was not provided. Please provide to proceed.")

    #         if forecast_period != testdata.shape[0]:
    #             warnings.warn("Forecast Period is not equal to the number of observations in testdata. The forecast period will be assumed to be the number of observations in testdata.")

    #         forecast_period = testdata.shape[0]

    #         try:
    #             testdata = testdata[self.original_preds]
    #         except Exception as e:
    #             print(e)
    #             raise ValueError("Some exogenous columns that were used during training are missing in testdata. Please make sure you are passing the correct exogenous columns.")

    #     if self.univariate:
    #         res = self.model.get_forecast(forecast_period)
    #     else:
    #         res = self.model.get_forecast(forecast_period, exog=testdata)

    #     res_frame = res.summary_frame()

    #     if simple:
    #         res_frame = res_frame['mean']
    #         res_frame = res_frame.squeeze() # Convert to a pandas series object
    #     else:
    #         # Pass as is
    #         pass

    #     return res_frame



import math
import pdb
import numpy as np  # type: ignore
import pandas as pd  # type: ignore

# TODO: Resolve which one we want to use
#from pmdarima.arima.auto import auto_arima # type: ignore
from pmdarima.arima import auto_arima # type: ignore


from .build_arima_base import BuildArimaBase

# helper functions
from ...utils import colorful, print_static_rmse, print_dynamic_rmse
from ...models.ar_based.param_finder import find_best_pdq_or_PDQ


class BuildAutoSarimax(BuildArimaBase):

    def find_best_parameters(self, data: pd.DataFrame):
        """
        Given a dataset, finds the best parameters using the settings in the class
        """

        if self.verbose >= 1:
            print(colorful.BOLD + '\n    Finding the best parameters using AutoArima:' + colorful.END)
        if len(self.original_preds) == 0:
            exog = None
        elif len(self.original_preds) == 1:
            exog = data[self.original_preds[0]].values.reshape(-1, 1)
        else:
            exog = data[self.original_preds].values

        ### for large datasets, speed is of the essence. Hence reduce max size of PDQ
        if data.shape[0] > 1000:
            print('    Using smaller parameters for larger dataset with greater than 1000 samples')
            out_of_sample_size = int(0.01*data.shape[0])
            arima_model = auto_arima(
                y = data[self.original_target_col],
                exogenous=exog, ## these variables must be given in predictions as well
                start_p = 0, start_q = 0, start_P = 0, start_Q = 0,
                max_p = 2, max_q = 2, max_P = 2, max_Q = 2,
                D = 1, max_D = 1,
                out_of_sample_size=out_of_sample_size,  # use a small amount
                information_criterion=self.scoring, # AIC
                scoring='mse', # only supports 'mse' or 'mae'
                m=self.seasonal_period, seasonal=self.seasonality,
                stepwise = True, random_state=42, n_fits = 10, n_jobs=-1,
                error_action = 'ignore')
        else:
            arima_model =  auto_arima(
                y = data[self.original_target_col],
                exogenous=exog, ## these variables must be given in predictions as well
                out_of_sample_size=0,  # use whole dataset to compute metrics
                information_criterion=self.scoring, # AIC
                scoring='mse', # only supports 'mse' or 'mae'
                # TODO: Check if we can go higher on max p and q (till seasonality)
                start_p=0, d=None, start_q=0, max_p=self.p_max, max_d=self.d_max, max_q=self.q_max, # AR Parameters
                start_P=0, D=None, start_Q=0, max_P=self.p_max, max_D=self.d_max, max_Q=self.q_max, # Seasonal Parameters (1)
                m=self.seasonal_period, seasonal=self.seasonality, # Seasonal Parameters (2)
                stepwise = True, random_state=42, n_fits = 50, n_jobs=-1,  # Hyperparameer Search
                error_action='warn', trace = True, supress_warnings=True
                    )

        self.best_p, self.best_d, self.best_q = arima_model.order  # example (0, 1, 1)
        self.best_P, self.best_D, self.best_Q, _ = arima_model.seasonal_order # example (2, 1, 1, 12)

        metric_value = math.nan

        if self.scoring.lower() == 'aic':
            metric_value = arima_model.aic()
        elif self.scoring.lower() == 'aicc':
            metric_value = arima_model.aicc()
        elif self.scoring.lower() == 'bic':
            metric_value = arima_model.bic()
        else:
            print("Error: Metric must be 'aic', 'aicc', or 'bic'. Continuing with 'bic' as default")
            metric_value = arima_model.bic()
            self.scoring = 'bic'

        if self.verbose >= 1:
            print(
                '\nBest model is a Seasonal SARIMAX(%d,%d,%d)*(%d,%d,%d,%d), %s = %0.3f' % (
                self.best_p, self.best_d, self.best_q,
                self.best_P, self.best_D, self.best_Q,
                self.seasonal_period, self.scoring, metric_value)
            )
        return arima_model



from typing import Optional
import warnings
warnings.filterwarnings(action='ignore')
from abc import abstractmethod
import copy
import pdb

import numpy as np  # type: ignore
import pandas as pd  # type: ignore
from pandas.core.generic import NDFrame # type:ignore

import matplotlib.pyplot as plt # type: ignore

from tscv import GapWalkForward # type: ignore

# imported SARIMAX from statsmodels pkg
from statsmodels.tsa.statespace.sarimax import SARIMAX  # type: ignore

from ..build_base import BuildBase

# helper functions
from ...utils import colorful, print_static_rmse, print_dynamic_rmse
from ...models.ar_based.param_finder import find_best_pdq_or_PDQ


class BuildArimaBase(BuildBase):
    def __init__(self, scoring, seasonality=False, seasonal_period=None, p_max=12,
            d_max=2, q_max=12, forecast_period=5, verbose=0):
        """
        Base class for building any ARIMA model
        Definitely applicable to SARIMAX and auto_arima with seasonality
        Check later if same can be reused for ARIMA (most likely yes)
        """
        super().__init__(
            scoring=scoring,
            forecast_period=forecast_period,
            verbose=verbose
        )

        self.seasonality = seasonality
        self.seasonal_period = seasonal_period
        self.p_max = p_max
        self.d_max = d_max
        self.q_max = q_max

        self.best_p = None
        self.best_d = None
        self.best_q = None
        self.best_P = None
        self.best_D = None
        self.best_Q = None


    def fit(self, ts_df: pd.DataFrame, target_col: str, cv: Optional[int]=None):
        """
        Build a Time Series Model using SARIMAX from statsmodels.
        """

        self.original_target_col = target_col
        self.original_preds = [x for x in list(ts_df) if x not in [self.original_target_col]]

        if len(self.original_preds) == 0:
            self.univariate = True
        else:
            self.univariate = False


        ##########################################
        #### Find best pdq and PDQ parameters ####
        ##########################################

        # NOTE: We use the entire dataset to compute the pdq and PDQ parameters.
        # Then we use the selected "best" parameters to check how well it
        # generalizes across the various folds (which may even be 1)

        # ## Added temporarily
        # ts_train = ts_df.iloc[:-self.forecast_period]
        # self.find_best_parameters(data = ts_train)
        auto_arima_model = self.find_best_parameters(data = ts_df)
        self.model = auto_arima_model

        if self.verbose >= 1:
            print(f"\n\nBest Parameters:")
            print(f"p: {self.best_p}, d: {self.best_d}, q: {self.best_q}")
            print(f"P: {self.best_P}, D: {self.best_D}, Q: {self.best_Q}")
            print(f"Seasonality: {self.seasonality}\nSeasonal Period: {self.seasonal_period}")

        #######################################
        #### Cross Validation across Folds ####
        #######################################

        rmse_folds = []
        norm_rmse_folds = []
        forecast_df_folds = []

        NFOLDS = self.get_num_folds_from_cv(cv)

        #########################################################################
        num_obs = ts_df.shape[0]
        if self.forecast_period <= 5:
            #### Set a minimum of 5 for the number of rows in test!
            self.forecast_period = 5
        ### In case the number of forecast_period is too high, just reduce it so it can fit into num_obs
        if NFOLDS*self.forecast_period > num_obs:
            self.forecast_period = int(num_obs/(NFOLDS+1))
            print('Lowering forecast period to %d to enable cross_validation' %self.forecast_period)
        #########################################################################

        cv = GapWalkForward(n_splits=NFOLDS, gap_size=0, test_size=self.forecast_period)
        for fold_number, (train, test) in enumerate(cv.split(ts_df)):
            ts_train = ts_df.iloc[train]
            ts_test = ts_df.iloc[test]

            if self.verbose >= 1:
                print(f"\n\nFold Number: {fold_number+1} --> Train Shape: {ts_train.shape} Test Shape: {ts_test.shape}")

            ### this is needed for static forecasts ####################
            # TODO: Check if this needs to be fixed to pick usimg self.original_target_col
            y_truth = ts_train[:]  #  TODO: Note that this is only univariate analysis

            if len(self.original_preds) == 0:
                exog = None
            elif len(self.original_preds) == 1:
                exog = ts_test[self.original_preds[0]].values.reshape(-1, 1)
            else:
                exog = ts_test[self.original_preds].values

            y_forecasted = self.model.predict(ts_test.shape[0],exog)

            concatenated = pd.DataFrame(np.c_[ts_test[self.original_target_col].values,
                            y_forecasted], columns=['original', 'predicted'],index=ts_test.index)

            ### for SARIMAX and Auto_ARIMA, you don't have to restore differences since it predicts like actuals.###
            y_true = concatenated['original']
            y_pred = concatenated['predicted']

            if self.verbose >= 1:
                print('Static Forecasts:')
                # Since you are differencing the data, some original data points will not be available
                # Hence taking from first available value.
                print_static_rmse(y_true, y_pred, verbose=self.verbose)
                #quick_ts_plot(y_true, y_pred)

            # Extract the dynamic predicted and true values of our time series
            forecast_df = copy.deepcopy(y_forecasted)
            forecast_df_folds.append(forecast_df)


            rmse, norm_rmse = print_static_rmse(y_true, y_pred, verbose=0) ## don't print this time
            rmse_folds.append(rmse)
            norm_rmse_folds.append(norm_rmse)

            # TODO: Convert rmse_folds, rmse_norm_folds, forecasts_folds into base class attributes
            # TODO: Add gettes and seters for these class attributes.
            # This will ensure consistency across various model build types.


        # This is taking the std of entire dataset and using that to normalize
        # vs. other approach that was using std of individual folds to standardize.
        # Technically this is not correct, but in order to do Apples:Aples compatison with ML
        # (sklearn) based cross_val_score, we need to do this since we dont get individual folds
        # back for cross_val_score. If at a later point in time, we can get this, then,
        # we can revert back to dividing by individual fold std values.
        norm_rmse_folds2 = rmse_folds/ts_df[self.original_target_col].values.std()  # Same as what was there in print_dynamic_rmse()

        print(f"\nSARIMAX RMSE (all folds): {np.mean(rmse_folds):.4f}")
        print(f"SARIMAX Norm RMSE (all folds): {(np.mean(norm_rmse_folds2)*100):.0f}%\n")

        ###############################################
        #### Refit the model on the entire dataset ####
        ###############################################
        self.refit(ts_df=ts_df)

        if self.verbose >= 1:
            print(self.model.summary())

        # return self.model, forecast_df_folds, rmse_folds, norm_rmse_folds
        return self.model, forecast_df_folds, rmse_folds, norm_rmse_folds2

    def refit(self, ts_df: pd.DataFrame) -> object:
        """
        Refits an already trained model using a new dataset
        Useful when fitting to the full data after testing with cross validation
        :param ts_df The time series data to be used for fitting the model
        :type ts_df pd.DataFrame
        :rtype object
        """

        bestmodel = self.get_best_model(ts_df)

        print(colorful.BOLD + 'Refitting data with previously found best parameters' + colorful.END)
        try:
            self.model = bestmodel.fit(disp=False)
            print('    Best %s metric = %0.1f' % (self.scoring, eval('self.model.' + self.scoring)))
        except Exception as e:
            print(e)

        return self

    @abstractmethod
    def find_best_parameters(self, data: pd.DataFrame):
        """
        Given a dataset, finds the best parameters using the settings in the class
        Need to set the following parameters in the child class
        self.best_p, self.best_d, self.best_q
        self.best_P, self.best_D, self.best_Q
        """



    def get_best_model(self, data: pd.DataFrame):
        """
        Returns the 'unfit' SARIMAX model with the given dataset and the
        selected best parameters. This can be used to fit or refit the model.
        """

        # In order to get forecasts to be in the same value ranges of the orig_endogs, you
        # must  set the simple_differencing = False and the start_params to be the same as ARIMA.
        # That is the only way to ensure that the output of this model iscomparable to other ARIMA models

        if not self.seasonality:
            if self.univariate:
                bestmodel = SARIMAX(
                    endog=data[self.original_target_col],
                    # exog=data[self.original_preds], ###if it is univariate, no preds needed
                    order=(self.best_p, self.best_d, self.best_q),
                    enforce_stationarity=False,
                    enforce_invertibility=False,
                    trend='ct',
                    start_params=[0, 0, 0, 1],
                    simple_differencing=False)
            else:
                bestmodel = SARIMAX(
                    endog=data[self.original_target_col],
                    exog=data[self.original_preds], ## if it is multivariate, preds are needed
                    order=(self.best_p, self.best_d, self.best_q),
                    enforce_stationarity=False,
                    enforce_invertibility=False,
                    trend='ct',
                    start_params=[0, 0, 0, 1],
                    simple_differencing=False)
        else:
            if self.univariate:
                bestmodel = SARIMAX(
                    endog=data[self.original_target_col],
                    # exog=data[self.original_preds], ### if univariate, no preds are needed
                    order=(self.best_p, self.best_d, self.best_q),
                    seasonal_order=(self.best_P, self.best_D, self.best_Q, self.seasonal_period),
                    enforce_stationarity=False,
                    enforce_invertibility=False,
                    trend='ct',
                    start_params=[0, 0, 0, 1],
                    simple_differencing=False
                )
            else:
                bestmodel = SARIMAX(
                    endog=data[self.original_target_col],
                    exog=data[self.original_preds], ### if multivariate, preds are needed
                    order=(self.best_p, self.best_d, self.best_q),
                    seasonal_order=(self.best_P, self.best_D, self.best_Q, self.seasonal_period),
                    enforce_stationarity=False,
                    enforce_invertibility=False,
                    trend='ct',
                    start_params=[0, 0, 0, 1],
                    simple_differencing=False
                )

        return bestmodel

    def predict(
        self,
        testdata: Optional[pd.DataFrame]=None,
        forecast_period: Optional[int] = None,
        simple: bool = True) -> NDFrame:
        """
        Return the predictions
        """
        # Extract the dynamic predicted and true values of our time series
        if self.univariate:
            if isinstance(testdata, pd.DataFrame) or isinstance(testdata, pd.Series):
                # use the forecast period used during training
                forecast_period = testdata.shape[0]
                self.forecast_period = testdata.shape[0]
        else:
            if testdata is None:
                raise ValueError("SARIMAX needs testdata to make predictions, but this was not provided. Please provide to proceed.")
                forecast_period = self.forecast_period
            elif isinstance(testdata, pd.DataFrame) or isinstance(testdata, pd.Series):
                if forecast_period != testdata.shape[0]:
                    warnings.warn("Forecast Period is not equal to the number of observations in testdata. The forecast period will be assumed to be the number of observations in testdata.")
                forecast_period = testdata.shape[0]
                self.forecast_period = forecast_period
                try:
                    testdata = testdata[self.original_preds]
                except Exception as e:
                    print(e)
                    print("Model was trained with train dataframe. Please make sure you are passing a test data frame.")
                    return
            elif isinstance(testdata, int):
                if forecast_period != testdata:
                    print("Forecast Period is not equal to the number of observations in testdata. The forecast period will be assumed to be the number of observations in testdata.")

                forecast_period = testdata
                self.forecast_period = forecast_period

        if self.univariate:
            res = self.model.get_forecast(self.forecast_period)
        else:
            if isinstance(testdata, pd.DataFrame) or isinstance(testdata, pd.Series):
                res = self.model.get_forecast(self.forecast_period, exog=testdata)
            else:
                try:
                    res = self.model.get_forecast(self.forecast_period)
                except Exception as e:
                    print(e)
                    print("Model was trained with train dataframe. Please make sure you are passing a test data frame.")
                    return

        res_frame = res.summary_frame()
        res_frame.rename(columns = {'mean':'yhat'}, inplace=True)

        if simple:
            res_frame = res_frame['yhat']
            res_frame = res_frame.squeeze() # Convert to a pandas series object
        else:
            # Pass as is
            pass

        return res_frame
